#include <cstdlib>
#include <queue>

#include "compmod.hpp"
#include "log.hpp"
#include "resmes.hpp"

using namespace std;

extern prolog_program *program;

map<string, set<int> > compmod::entry::all_possibles;

void compmod::finalise()
{
   lout(L_STATUS) << "Finalising computation model..." << endl;
   // Delete the rest of the dynamically allocated memory
   deep_delete();
   for(size_t i = 0; i < program->rules.size(); i++)
      program->rules[i]->deep_delete();
   delete program;
}

void compmod::fatal_error() const
{
   throw compmod_error();
}

compmod::strategy compmod::get_strategy(const string &name)
{
   if(name == "first")
      return STRATEGY_FIRST;
   else if(name == "last")
      return STRATEGY_LAST;
   else
      return STRATEGY_RANDOM;
}

void compmod::report_answer(const c_term *ans_head) const
{
   con_type con;
   if(get_answer(ans_head, con))
   {
      if(con.is_false())
	 lout(L_REPORT) << ans_head << " is never satisfied." << endl;
      else if(con.is_true())
	 lout(L_REPORT) << ans_head << " is always satisfied." << endl;
      else
      {
	 lout(L_REPORT) << ans_head << " is satisfied under the following conditions:" << endl;
	 conaux::report_answer(con);
      }
   }
   else
      lout(L_REPORT) << "No answer calculated for " << ans_head << endl;
}

compmod::entry::entry(const c_rule *r): rule(r)
{
   if(rule == 0)
      return;

   if(rule->body.size() > 0)
   {
      const c_term *t = rule->body[0];
      string id = t->name;
      if(t->arglist.size() > 0 && t->arglist[0]->type == T_FUNCTOR)
	 id += "_" + t->arglist[0]->name;

      lout(L_DBG) << "Looking up possibles for " << id << endl;
      possibles = all_possibles[id];
      lout(L_DBG) << "Found possibles: ";
      lout(L_DBG).log_all(possibles, " ");
      lout(L_DBG) << endl;
   }
}

void compmod::entry::init()
{
   lout(L_STATUS) << "Initialising entries..." << endl;
   all_possibles.clear();

   for(size_t i = 0; i < program->rules.size(); i++)
   {
      const c_term *t = program->rules[i]->head;
      string id = t->name;
      if(t->arglist.size() > 0 && t->arglist[0]->type == T_FUNCTOR)
	 id += "_" + t->arglist[0]->name;
      
      all_possibles[id].insert(i);
   }

   map<string, set<int> >::iterator it = all_possibles.begin();
   while(it != all_possibles.end())
   {
      lout(L_DBG) << "Possibles for " << it->first << ": { ";
      lout(L_DBG).log_all(it->second, " ");
      lout(L_DBG) << "}" << endl;
      ++it;
   }
}

bool compmod::entry::operator<(const compmod::entry &rhs) const
{
   return c_sorter::c_rule_lt(rule, rhs.rule);
}

compmod::compmod(c_rule *rule, string &random_seed)
{
   resources["unify"].set_desc("Unification");
   resources["update"].set_desc("Update mechanism");
   resources["ebc"].set_desc("EBC/EBCWA");
   resources["ebaa"].set_desc("EBAA");
   resources["ebab"].set_desc("EBAB");
   resources["conproj"].set_desc("Constraint projection");
   resources["conop"].set_desc("Constraint macro operations");

   calls.insert(rule);
   if(random_seed == "random")
   {
      time_t r = time(0);
      ostringstream oss;
      oss << r;
      random_seed = oss.str();
      srand(r);
   }
   else
   {
      time_t r;
      istringstream iss(random_seed);
      iss >> r;
      srand(r);
   }  
}

void compmod::deep_delete()
{
   for(size_t i = 0; i < all_rules.size(); i++)
      all_rules[i]->deep_delete();

   map<const c_rule*, pair<const c_rule*, unifier*> >::const_iterator mapit = answer_extension.begin();
   while(mapit != answer_extension.end())
   {
      delete mapit->second.second;
      ++mapit;
   }
}

// static
compmod::entry compmod::get_next_entry(const set<entry> &s, compmod::strategy pick_strategy) const
{
   int idx;

   switch(pick_strategy)
   {
      case STRATEGY_FIRST:
	 idx = 0;
	 break;
	 	 
      case STRATEGY_LAST:
	 idx = -1;
	 break;

      case STRATEGY_RANDOM:
      default:
	 idx = rand() % s.size(); // range [0, size - 1]
         break;
   }   

   if(idx == -1)
      return *(--s.end());
   else
   {
      set<entry>::const_iterator it = s.begin();
      advance(it, idx);
      return *it;
   }
}

compmod::entry compmod::get_next_answer_alternative() const
{
   return get_next_entry(ans_a, ans_a_pick_strategy);
}

compmod::entry compmod::get_next_answer_bound() const
{
   return get_next_entry(ans_b, ans_b_pick_strategy);
}

compmod::entry compmod::get_next_call() const
{
   return get_next_entry(calls, calls_pick_strategy);
}

bool compmod::extend_set()
{
   size_t tot_amount = calls.size() + ans_a.size() + ans_b.size();

   if(tot_amount == 0)
      return false; // None available
   
   if(pick_strategy == "weighted")
   {
      size_t idx = rand() % tot_amount;
      
      if(idx < calls.size())
      {
	 resources["ebc"].start();
	 extend_by_call();
	 resources["ebc"].stop();
         return true;
      }
      else if(idx - calls.size() < ans_a.size())
      {
	 resources["ebaa"].start();
	 extend_by_answer_alternative();
	 resources["ebaa"].stop();
         return true;
      }
      else
      {
	 resources["ebab"].start();
	 extend_by_answer_bound();
	 resources["ebab"].stop();
         return true;
      }
   }
   else // A permutation of abc
   {
      for(size_t i = 0; i < 3; i++)
      {
	 if(pick_strategy[i] == 'a' && ans_a.size() > 0)
	 {
	    resources["ebaa"].start();
	    extend_by_answer_alternative();
	    resources["ebaa"].stop();
            return true;
	 }
	 else if(pick_strategy[i] == 'b' && ans_b.size() > 0)
	 {
	    resources["ebab"].start();
	    extend_by_answer_bound();
	    resources["ebab"].stop();
            return true;
	 }
	 else if(pick_strategy[i] == 'c' && calls.size() > 0)
	 {
	    resources["ebc"].start();
	    extend_by_call();
	    resources["ebc"].stop();
            return true;
	 }
      }
   }

   return true;
}

void compmod::extend_by_call()
{
   entry call = get_next_call();

   lout(L_EXT) << "Got call: " << call << endl;

   // Decide whether to use local or global extension strategy
   bool local_extension;
   if(call.rule->body.size() == 1) // Answers should never be derived in a context
      local_extension = false;
   else if(percent_local_extensions >= 100)
      local_extension = true;
   else if(percent_local_extensions <= 0)
      local_extension = false;
   else
   {
      size_t temp = (rand() % 100) + 1; // range [1,100]
      if(temp <= percent_local_extensions)
	 local_extension = true;
      else
	 local_extension = false;
   }

   lout(L_EXT) << "Local extension? " << (local_extension?"YES":"NO") << endl;

   // If there are no matching rules in the program, remove call or extend according to CWA
   if(call.possibles.size() == 0)
   {
      // If this call has not been used for EBC, use it for EBCWA
      if(call.rule->body[0]->negated
	 && used_ebc.find(call.rule) == used_ebc.end())
      {
	 extend_by_first_answer(call.rule, con_type(false));
      }

      // Make this call obsolete
      calls.erase(call);
      ob_calls.insert(call);
      return;
   }
   
   // Get an index of a (possibly) matching rule from the original program
   size_t index;
   if(possibles_strategy == STRATEGY_FIRST)
     index = *call.possibles.begin();
   else if(possibles_strategy == STRATEGY_LAST)
     index = *(--call.possibles.end());
   else // possibles_STRATEGY == STRATEGY_RANDOM
   {
      index = rand() % call.possibles.size(); // range [0, size-1]
      set<int>::iterator it = call.possibles.begin();
      advance(it, index);
      index = *it;
   }

   set<entry>::iterator it2 = calls.find(call);
   it2->possibles.erase(index);
   if(it2->possibles.empty())
   {
      entry temp = *it2;
      ob_calls.insert(temp);
      calls.erase(temp);
   }

   lout(L_EXT) << "Index: " << index << " " << program->rules[index] << endl;

   // Get the (possibly) matching rule and try to unify
   unifier mgu;
   c_rule *rule_to_add = new c_rule;
   program->rules[index]->deep_copy(rule_to_add, true);
   
   resources["unify"].start();
   bool temp = mgu.unify(call.rule->body[0], rule_to_add->head);
   resources["unify"].stop();

   if(!temp)
   {
      rule_to_add->deep_delete();
      return;
   }

   // When extending locally, concatenate the constraint from the "parent rule".
   // We know that rule_to_add->constraint has the correct variable names.
   if(local_extension)
      rule_to_add->constraint = rule_to_add->constraint & call.rule->constraint;

   // Skip unsatisfiable constraints
   //if(rule_to_add->constraint.is_false())
   //{
   //   lout(L_EXT) << "***Constraint unsatisfiable, skipping!" << endl;
   //   rule_to_add->deep_delete();
   //   return;
   //}

   lout(L_DBG) << "mgu: " << &mgu << endl;

   // Apply the unifier
   mgu.apply_to(rule_to_add);

   // Add the derived entry to the set (and delete the unifier)
   pair<const c_rule*, unifier*> p = add_to_set(rule_to_add);
   delete p.second;

   // Mark that this call tuple has been used for EBC
   // (which would prevent EBCWA for this call tuple)
   used_ebc.insert(call.rule);
}

pair<const c_rule*, unifier*> compmod::add_to_set(c_rule *rule)
{
   lout(L_EXT) << "Trying to add to set: " << rule << endl;

   if(rule->body.size() == 0) // It's an answer alternative
   {
      set<entry>::iterator setit = ans_a.find(rule);
      unifier *mgu = new unifier;

      if(setit == ans_a.end()) // There was no unused answer alternative there before
      {
	 lout(L_EXT) << "Adding answer" << endl;
         rule->project_constraint();
	 ans_a.insert(rule);
	 all_rules.push_back(rule);
	 return pair<const c_rule*, unifier*>(rule, mgu);
      }
      else
      {
	 lout(L_EXT) << "Updating answer" << endl;
         rule->project_constraint();
	 con_type con = rule->constraint;
	 resources["unify"].start();
	 mgu->unify(setit->rule, rule);
	 resources["unify"].stop();
	 mgu->apply_to(con);
	 setit->rule->constraint = setit->rule->constraint | con;
         setit->rule->project_constraint();
	 rule->deep_delete();
	 return pair<const c_rule*, unifier*>(setit->rule, mgu);
      }
   }
   else // It's a call
   {
      set<entry>::iterator setit;
      entry ent(rule);

      if((setit = ob_calls.find(ent)) != ob_calls.end()
	 || (setit = calls.find(ent)) != calls.end())
      {
	 lout(L_EXT) << "Updating call: " << rule << endl;
	 unifier *mgu = new unifier;
	 resources["unify"].start();
	 bool temp = mgu->unify(setit->rule, rule);
	 resources["unify"].stop();
	 if(!temp)
	 {
	    lerr(L_ERR) << "***Conflicting rules (should not happen): " << endl
			<< setit->rule << endl
			<< rule << endl;
	    fatal_error();
	 }

         rule->project_constraint();
	 mgu->apply_to(rule->constraint);
	 change_constraint_disjunction(setit->rule, rule->constraint);
     
	 rule->deep_delete();
	 return pair<const c_rule*, unifier*>(setit->rule, mgu);
      }
      else
      {
         rule->project_constraint();
	 lout(L_EXT) << "Adding call: " << rule << endl;
	 calls.insert(ent);

	 // We need to check for already processed answers here and
	 // apply them directly.  Otherwise, answer updating will not
	 // work as expected later.
	 con_type old_answer;
	 lout(L_EXT) << "Checking for old answer to " << rule->body[0] << endl;
	 if(get_answer(rule->body[0], old_answer))
	 {	    
	    lout(L_EXT) << "Extending by old answer" << endl;
	    extend_by_first_answer(rule, old_answer);
	 }
	 all_rules.push_back(rule);
      }
   }

   return pair<const c_rule*, unifier*>(rule, new unifier); // No unifier
}

void compmod::get_calls(const c_term *t, vector<const c_rule*> &calls_ret) const
{
   int tempsize = t->arglist.size();

   c_rule *low = new c_rule;
   c_rule *high = new c_rule;

   c_term *low_first = new c_term;
   c_term *high_first = new c_term;

   t->deep_copy(high_first);
   t->deep_copy(low_first);

   low_first->arglist[tempsize - 1]->type = -1000;
   high_first->arglist[tempsize - 1]->type = 1000;
   low->body.push_back(low_first);
   high->body.push_back(high_first);

   set<entry>::iterator low_it;
   set<entry>::iterator high_it;

   low_it = calls.lower_bound(entry(low));
   high_it = calls.upper_bound(entry(high));

   while(low_it != high_it)
   {
      calls_ret.push_back(low_it->rule);
      ++low_it;
   }

   low_it = ob_calls.lower_bound(entry(low));
   high_it = ob_calls.upper_bound(entry(high));

   while(low_it != high_it)
   {
      calls_ret.push_back(low_it->rule);
      ++low_it;
   }

   low->deep_delete();
   high->deep_delete();
}

void compmod::extend_by_answer_alternative()
{
   entry ans = get_next_answer_alternative();

   lout(L_EXT) << "Got answer alternative: " << ans << endl;

   bool first_time = (ob_ans_a.find(ans) == ob_ans_a.end());

   if(first_time)
   {
      // Mark this answer as obsolete
      ans_a.erase(ans);
      ob_ans_a.insert(ans); 

      vector<const c_rule*> calls;

      get_calls(ans.rule->head, calls);

      lout(L_DBG) << "Fitting calls to process:" << endl;
      lout(L_DBG).log_all(calls, "\n");
      lout(L_DBG) << endl;

      // Process all the fitting calls
      for(size_t i = 0; i < calls.size(); i++)
      {
	 con_type new_c = ans.rule->constraint;

         set<entry>::iterator it = ob_ans_b.find(ans);
         if(it != ob_ans_b.end()) // We have to adhere to previous bounds
         {
            lout(L_EXT) << "Adhering by previous bound" << endl;
            con_type old_c = it->rule->constraint;
            unifier mgu;
            resources["unify"].start();
            bool temp = mgu.unify(ans.rule, it->rule);
            resources["unify"].stop();
            if(!temp)
               return;
            
            mgu.apply_to(old_c);
            new_c = new_c & old_c; // Applying the old bound
         }

	 unifier mgu;
	 resources["unify"].start();
	 mgu.unify(calls[i]->body[0], ans.rule->head);
	 resources["unify"].stop();
	 mgu.apply_to(new_c);
	 extend_by_first_answer(calls[i], new_c);
      }
   }
   else
   {
      // The answer for this literal was used before, only perform the
      // answer update part

      // If the previous answers subsume this one, abort EBAA
      set<entry>::iterator it = ob_ans_a.find(ans);
      unifier ex_mgu;
      resources["unify"].start();
      ex_mgu.unify(ans.rule->head, it->rule->head);
      resources["unify"].stop();
      con_type ex_c = it->rule->constraint;
      ex_mgu.apply_to(ex_c);
      if(conaux::subsumes(ex_c, ans.rule->constraint))
      {
	 lout(L_EXT) << "Previous answer subsumed this one." << endl << ex_c << ans.rule->constraint;
	 ans_a.erase(ans);
	 return;
      }

      // We know this new answer alternative might contribute
      it = ob_ans_b.find(ans);
      con_type new_c = ans.rule->constraint;

      lout(L_EXT) << "Extending by answer alternative" << endl;

      if(it != ob_ans_b.end()) // We have to adhere to previous bounds
      {
	 lout(L_EXT) << "Adhering by previous bound" << endl;
	 con_type old_c = it->rule->constraint;
	 unifier mgu;
	 resources["unify"].start();
	 bool temp = mgu.unify(ans.rule, it->rule);
	 resources["unify"].stop();
	 if(!temp)
	    return;

	 mgu.apply_to(old_c);
	 new_c = new_c & old_c; // Applying the old bound
      }
      
      // Find all calls with the answer's head as their next body literal
      vector<const c_rule*> calls_vec;
      get_calls(ans.rule->head, calls_vec);

      //cout << endl << "Ans: " << ans.rule->head << endl << endl;
      lout(L_EXT) << "Fitting calls to update from:" << endl;
      lout(L_EXT).log_all(calls_vec, "\n");
      lout(L_EXT) << endl;

      for(size_t i = 0; i < calls_vec.size(); i++)
      {
	 unifier mgu;
	 resources["unify"].start();
	 bool temp_bool = mgu.unify(calls_vec[i]->body[0], ans.rule->head);
	 resources["unify"].stop();
	 if(!temp_bool)
	    continue;

	 con_type temp = new_c;
	 mgu.apply_to(temp);

	 pair<const c_rule*, unifier*> &p = answer_extension[calls_vec[i]];

	 if(calls_vec[i]->body[0]->negated) // negated
	 {
	    p.second->apply_to(temp);
	    change_constraint_conjunction(p.first, !temp);
	 }
	 else // non-negated
	 {
	    temp = temp & calls_vec[i]->constraint;
	    p.second->apply_to(temp);
	    change_constraint_disjunction(p.first, temp);
	 }
      }

      // Mark this answer as obsolete
      ans_a.erase(ans);
      it = ob_ans_a.find(ans);
      unifier mgu;
      resources["unify"].start();
      mgu.unify(it->rule, ans.rule);
      resources["unify"].stop();
      mgu.apply_to(ans.rule->constraint);
      it->rule->constraint = it->rule->constraint | ans.rule->constraint;
      it->rule->project_constraint();
   }
}

void compmod::extend_by_answer_bound()
{
   entry ans = get_next_answer_bound();

   lout(L_EXT) << "Got answer bound: " << ans << endl;
   
   set<entry>::iterator it = ob_ans_b.find(ans);

   // If this answer bound subsume a previous answer bound, abort EBAB
   if(it != ob_ans_b.end())
   {
      unifier ex_mgu;
      resources["unify"].start();
      ex_mgu.unify(ans.rule, it->rule);
      resources["unify"].stop();
      con_type ex_c = it->rule->constraint;
      ex_mgu.apply_to(ex_c);
      if(conaux::subsumes(ans.rule->constraint, ex_c))
      {
	 lout(L_EXT) << "This answer subsumed this previous one." << endl << ex_c << ans.rule->constraint;
	 ans_b.erase(ans);
	 return;
      }
   }

   // Find all calls with the answer's head as their next body literal
   vector<const c_rule*> calls_vec;
   get_calls(ans.rule->head, calls_vec);
   
   lout(L_DBG) << "Fitting calls to update from:" << endl;
   lout(L_DBG).log_all(calls_vec, "\n");
   lout(L_DBG) << endl;

   for(size_t i = 0; i < calls_vec.size(); i++)
   {
      unifier mgu;
      lout(L_DBG) << "Trying to unify (" << calls_vec[i]->body[0] << " with " << ans.rule->head << ")" << endl;
      resources["unify"].start();
      bool temp_bool = mgu.unify(calls_vec[i]->body[0], ans.rule->head);
      resources["unify"].stop();
      lout(L_DBG) << (temp_bool?"Unified":"Did not unify") << endl;
      if(!temp_bool)
	 continue;

      con_type temp = ans.rule->constraint;
      mgu.apply_to(temp);

      pair<const c_rule*, unifier*> &p = answer_extension[calls_vec[i]];

      lout(L_DBG) << "Updating " << p.first << endl;
      
      if(calls_vec[i]->body[0]->negated) // negated
      {
	 temp = calls_vec[i]->constraint & !temp;
	 p.second->apply_to(temp);
	 change_constraint_disjunction(p.first, temp);
      }
      else // non-negated
      {
	 p.second->apply_to(temp);
	 change_constraint_conjunction(p.first, temp);
      }
   }  

   // Remove it from the unused bounds and merge it with the used bounds (or insert it)
   ans_b.erase(ans);
   if(it == ob_ans_b.end())
      ob_ans_b.insert(ans);
   else
   {
      unifier mgu;
      resources["unify"].start();
      mgu.unify(it->rule, ans.rule);
      resources["unify"].stop();
      mgu.apply_to(ans.rule->constraint);
      it->rule->constraint = it->rule->constraint & ans.rule->constraint;
      it->rule->project_constraint();
   }
}

void compmod::extend_by_first_answer(const c_rule *rule, const con_type &answer)
{
   lout(L_EXT) << "Extending by first answer" << endl;

   // Copy the old rule
   c_rule *rule_to_add = new c_rule;
   rule->deep_copy(rule_to_add);

   // Erase and delete the first literal in the body
   c_term *first = rule_to_add->body[0];
   rule_to_add->body.erase(rule_to_add->body.begin(), rule_to_add->body.begin() + 1);
   first->deep_delete();

   // Concatenate the answer constraint to the rule we are adding (negated if it's a negative extension)
   if(rule->body[0]->negated)
      rule_to_add->constraint = rule_to_add->constraint & !answer; 
   else
      rule_to_add->constraint = rule_to_add->constraint & answer;  

   // Do not add this extension if the constraint is not satifiable
   // and we cannot extend the rule further.  This is surely the case
   // if the rule to be added has a body that contains positive
   // literals.
   
   // Project the constraint
   rule_to_add->project_constraint();

   // Mark the added (or already existing) rule as the answer
   // extension
   pair<const c_rule*, unifier*> p = add_to_set(rule_to_add);
   answer_extension[rule] = p;
}

// Changing an answer from "a :- C1" to "a :- C1 v C2"
void compmod::change_constraint_disjunction_answer(const c_rule *original, con_type addition)
{
   set<entry>::iterator it = ans_a.find(original);
   if(it == ans_a.end())
   {
      c_rule *rule_to_add = new c_rule;
      original->deep_copy(rule_to_add);
      rule_to_add->constraint = rule_to_add->constraint | addition;
      ans_a.insert(rule_to_add);
      all_rules.push_back(rule_to_add);
   }
   else
   {
      unifier mgu;
      resources["unify"].start();
      mgu.unify(it->rule, original);
      resources["unify"].stop();
      con_type c = addition;
      mgu.apply_to(c);
      it->rule->constraint = it->rule->constraint | c;
   }   
}

// Changing a call from "a :- C1, b..." to "a :- C1 v C2, b..."
void compmod::change_constraint_disjunction_call(const c_rule *original, con_type addition)
{
   // Only continue if something will be altered
   if(conaux::subsumes(original->constraint, addition))
      return;

   map<const c_rule*, pair<const c_rule*, unifier*> >::iterator it = answer_extension.find(original);
   if(it == answer_extension.end()) return;

   // First, we need to fetch the answer constraint for the next body literal
   con_type ans_constraint;
   unifier mgu;
   get_answer(original->body[0], ans_constraint);

   // Then we translate our addition to the right variables and project.
   con_type temp = addition;
   it->second.second->apply_to(temp);
   set<string> cvars;
   it->second.first->get_variables(cvars);
   addition.project(cvars);

   if(original->body[0]->negated) // Extension by negative answer
   {
      con_type c = temp & !ans_constraint;
      change_constraint_disjunction(it->second.first, c);
   }
   else // Extension by positive answer
   {
      con_type c = temp & ans_constraint;
      change_constraint_disjunction(it->second.first, c);
   }
}

// Changing an entry from "a :- C1, ..." to "a :- C1 v C2, ..."
void compmod::change_constraint_disjunction(const c_rule *original, con_type addition)
{
   if(original->body.size() == 0) // We are modifying an answer
   {
      resources["update"].start();
      change_constraint_disjunction_answer(original, addition);
      resources["update"].stop();
   }
   else // We are modifying a call
   {
      resources["update"].start();
      set<string> cvars;
      original->get_variables(cvars);
      addition.project(cvars);
      change_constraint_disjunction_call(original, addition);
      resources["update"].stop();
      
      // Apply the change to the rule itself
      original->constraint = original->constraint | addition;
   }
   
   // Project the constraint
   original->project_constraint();
}

// Changing an answer from "a :- C1" to "a :- C1 ^ C2"
void compmod::change_constraint_conjunction_answer(const c_rule *original, con_type addition)
{
   set<string> cvars;
   original->get_variables(cvars);
   addition.project(cvars);

   lout(L_DBG) << "Trying to add answer bound " << original << endl << bddtable << addition << endl;

   set<entry>::iterator it = ans_b.find(original);
   if(it == ans_b.end()) // There was no bound in ans_b for this head
   {
      /* We want to maintain the property that each time an EBAB is
       * performed, an earlier EBAA for the same head has been
       * performed.
       *
       * Because of this, we cannot simply add this bound unless a
       * corresponding head exists in ob_ans_a.
       *
       * But since used bounds are checked before every EBAA
       * application, we are allowed to just put this bound in
       * ob_ans_b (if no earlier EBAA for this head has been
       * made). The end result would be the same.
       */
      
      it = ob_ans_a.find(original);
      if(it != ob_ans_a.end()) // A corresponding answer alternative
                               // has been used, add the bound to
                               // ans_b
      {
	 c_rule *rule_to_add = new c_rule;
	 original->deep_copy(rule_to_add);
	 rule_to_add->constraint = rule_to_add->constraint & addition;
	 ans_b.insert(rule_to_add);
	 all_rules.push_back(rule_to_add);
      }
      else // No corresponding answer alternative has been used, add
           // or concatenate this bound to ob_ans_b
      {
	 it = ob_ans_b.find(original);
	 if(it == ob_ans_b.end()) // No such bound, add this one (*)
	 {
	    c_rule *rule_to_add = new c_rule;
	    original->deep_copy(rule_to_add);
	    rule_to_add->constraint = addition;
	    ob_ans_b.insert(rule_to_add);
	    all_rules.push_back(rule_to_add);
	 }
	 else // Translate to and concatenate with the previous bound (was added by (*))
	 {
	    unifier mgu;
	    resources["unify"].start();
	    mgu.unify(it->rule, original);
	    resources["unify"].stop();

	    con_type c = addition;
	    mgu.apply_to(c);
	    it->rule->constraint = it->rule->constraint & c;
	 }
      }
   }
   else // There was already a bound in ans_b for this head, concatenate with it
   {
      unifier mgu; 
      resources["unify"].start();
      mgu.unify(it->rule, original);
      resources["unify"].stop();

      con_type c = addition;
      mgu.apply_to(c); 
      it->rule->constraint = it->rule->constraint & c;
   }
}

// Changing a call from "a :- C1, b..." to "a :- C1 ^ C2, b..."
void compmod::change_constraint_conjunction_call(const c_rule *original, con_type addition)
{
   // Only continue if something will be altered
   if(conaux::subsumes(original->constraint, addition))
      return;

   map<const c_rule*, pair<const c_rule*, unifier*> >::iterator it = answer_extension.find(original);

   if(it == answer_extension.end()) return;

   it->second.second->apply_to(addition);
   set<string> cvars;
   it->second.first->get_variables(cvars);
   addition.project(cvars);
   
   change_constraint_conjunction(it->second.first, addition);
}

// Changing an entry from "a :- C1, ..." to "a :- C1 ^ C2, ..."
void compmod::change_constraint_conjunction(const c_rule *original, con_type addition)
{
   if(original->body.size() == 0) // We are modifying an answer
   {
      resources["update"].start();
      change_constraint_conjunction_answer(original, addition);
      resources["update"].stop();
   }
   else // We are modifying a call
   {
      resources["update"].start();
      set<string> cvars;
      original->get_variables(cvars);
      addition.project(cvars);
      change_constraint_conjunction_call(original, addition);
      resources["update"].stop();

      // Apply the change to the rule itself
      original->constraint = original->constraint & addition;
   }

   // Project the constraint
   original->project_constraint();
}

// The returned answer has the same variable names as the term argument
bool compmod::get_answer(const c_term *t, con_type &answer) const
{
   c_rule *ans_rule = new c_rule;
   t->deep_copy(ans_rule->head);
   // We only seek for answers to positive literals, worry about negation later
   ans_rule->head->negated = false;

   set<entry>::const_iterator ansit_a = ob_ans_a.find(ans_rule);
   set<entry>::const_iterator ansit_b = ob_ans_b.find(ans_rule);

   con_type ans_a, ans_b;

   ans_rule->deep_delete();

   // There will always be a used answer alternative if there is a used answer at all
   if(ansit_a == ob_ans_a.end())
      return false;
 
   unifier mgu;
   resources["unify"].start();
   bool temp = mgu.unify(t, ansit_a->rule->head);
   resources["unify"].stop();

   // Get the unifier of the answer and the term we are seeking the answer for
   if(!temp)
   {
      lerr(L_ERR) << "Bad alternative answer present." << endl
		  << "alt: " << ansit_a->rule << endl;
      fatal_error();
   }

   ans_a = ansit_a->rule->constraint;
   mgu.apply_to(ans_a);

   if(ansit_b != ob_ans_b.end())
   {
      unifier mgu2; // The structural bindings will be the same as in mgu
      resources["unify"].start();
      bool temp = mgu2.unify(t, ansit_b->rule->head);
      resources["unify"].stop();
      if(!temp)
      {
	 lerr(L_ERR) << "Bad alternative bound present." << endl
		     << "  alt: " << ansit_a->rule << endl
		     << "bound: " << ansit_b->rule << endl;
	 fatal_error();
      }
      ans_b = ansit_b->rule->constraint;
      mgu2.apply_to(ans_b);
   }
   else
      ans_b = true;

   answer = ans_a & ans_b;
   
   set<string> cvars;
   t->get_variables(cvars);
   answer.project(cvars);

   return true;
}

void compmod::set_pick_strategy(const string &s)
{
   if((s.size() == 3 && (s.find('a') == string::npos
                         || s.find('b') == string::npos
                         || s.find('c') == string::npos))
      || (s.size() != 3 && s != "weighted"))
   {
      throw runtime_error("Invalid strategy '" + s + "' chosen.");
   }
   else
      pick_strategy = s;
}

void compmod::set_ans_a_pick_strategy(const strategy &s)
{
   ans_a_pick_strategy = s;
}

void compmod::set_ans_b_pick_strategy(const strategy &s)
{
   ans_b_pick_strategy = s;
}

void compmod::set_calls_pick_strategy(const strategy &s)
{
   calls_pick_strategy = s;
}

void compmod::set_possibles_strategy(const strategy &s)
{
   possibles_strategy = s;
}

void compmod::set_local_ebc_percentage(size_t p)
{
   percent_local_extensions = p;
}

ostream &operator<<(ostream &os, const compmod::entry &e)
{
   os << e.rule << " { ";

   set<int>::const_iterator it = e.possibles.begin();

   if(it != e.possibles.end())
   {
      os << *it;
      ++it;
   }

   while(it != e.possibles.end())
   {
      os << ", " << *it;
      ++it;
   }

   os << " }" << endl << e.rule->constraint;

   return os;
}

ostream &operator<<(ostream &os, const compmod &c)
{
   os << "--Answer alternatives (" << c.ans_a.size() << ")" << endl;
   copy(c.ans_a.begin(), c.ans_a.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   os << "--Answer bounds (" << c.ans_b.size() << ")" << endl;
   copy(c.ans_b.begin(), c.ans_b.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   os << "--Calls (" << c.calls.size() << ")" << endl;
   copy(c.calls.begin(), c.calls.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   os << "--Obsolete answer alternatives (" << c.ob_ans_a.size() << ")" << endl;
   copy(c.ob_ans_a.begin(), c.ob_ans_a.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   os << "--Obsolete answer bounds (" << c.ob_ans_b.size() << ")" << endl;
   copy(c.ob_ans_b.begin(), c.ob_ans_b.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   os << "--Obsolete calls (" << c.ob_calls.size() << ")" << endl;
   copy(c.ob_calls.begin(), c.ob_calls.end(), ostream_iterator<compmod::entry>(os, ""));
   os << endl;

   return os;
}
