#ifndef COMPMOD_H
#define COMPMOD_H

#include <vector>
#include <cstdlib>
#include <iterator>
#include <set>
#include <map>
#include <stdexcept>

#include "parser-classes.hpp"
#include "unify.hpp"

/**
 * This class represents the computation model and encapsulates all
 * operations that can be performed on such a model. This is the main
 * class used in the calculation.
 */
class compmod
{
public:

   /**
    * The values possible for the picking strategy.
    */
   enum strategy { STRATEGY_RANDOM = 1, STRATEGY_FIRST,
                   STRATEGY_LAST };
   
   /**
    * An exception class denoting a serious error in the computation model.
    */
   class compmod_error: public std::exception {};
   
   /**
    * An entry in the computation set.
    */
   class entry
   {
   public:

      /**
       * Write a text representation of this entry to an ostream.
       * \param os the ostream to write to.
       * \param e the entry to write.
       */
      friend std::ostream &operator<<(std::ostream &os, const entry &e);

      /**
       * Create an entry based on the specified c_rule.
       * \param r the clause to base the entry on.
       */
      entry(const c_rule *r);

      /**
       * Initialises the entry class. This precalculates all
       * sets of possible extensions of call entries.
       */
      static void init();

      /**
       * Determine whether this entry is to be ordered before
       * another.
       * \param rhs the other entry. 
       * \return Truth value.
       */
      bool operator<(const entry &rhs) const;

      /**
       * The rule this entry is based on.
       */
      const c_rule *rule;

      /**
       * The set of possible extensions of this entry (only
       * applicable for calls).
       */
      mutable std::set<int> possibles;

   private:

      /**
       * All possible sets of possible extensions of entries,
       * indexed on prefix.
       */
      static std::map<std::string, std::set<int> > all_possibles;
   };

   /**
    * Write a text representation of a computation model's state to
    * an ostream.
    * \param os the ostream to write to.  
    * \param c the model, whose state to write.
    */
   friend std::ostream &operator<<(std::ostream &os, const compmod &c);

   /**
    * Constructor that initialises the random number generator,
    * inserts the query in the model and sets the pick strategies.
    * \param rule the query to solve.
    * \param random_seed the random seed to use (or "random").
    */
   compmod(c_rule *rule, std::string &random_seed);
      
   /**
    * Finalise the computation model. This clears all used dynamic memory.
    */
   void finalise();

   /**
    * Translates a strategy name into a strategy token.
    * \param name the strategy name.
    * \return The corresponding strategy.
    */
   static strategy get_strategy(const std::string &name);

   /**
    * Reports the answer for the given head (usually the given query).
    * \param ans_head the head of the answer.
    */
   void report_answer(const c_term *ans_head) const;

   /**
    * Extends the model by picking an entry (answer alternative or bound or call),
    * depending on the choosing strategy and using the appropriate extension operation.
    * \return False if we have reached a fixpoint. True if an extension was tried.
    */
   bool extend_set();

   /**
    * Sets the global strategy for choosing whether to use an answer alternative,
    * answer bound or a call when extending the model.
    * \param s either "weighted", in which case a weighted random choice is made,
    * or a permutation of the letters "ABC", in which case answer alternatives (A),
    * answer bounds (B) or calls (C) will be preferred first. "ABC" means that we pick
    * answer alternatives if they exist, otherwise answer bounds if they exist and
    * lastly calls.
    */
   void set_pick_strategy(const std::string &s);

   /**
    * Sets the strategy for choosing which answer alternative to use if several exist.
    * \param s one of STRATEGY_FIRST, STRATEGY_LAST and STRATEGY_RANDOM.
    */
   void set_ans_a_pick_strategy(const strategy &s);

   /**
    * Sets the strategy for choosing which answer bound to use if several exist.
    * \param s one of STRATEGY_FIRST, STRATEGY_LAST and STRATEGY_RANDOM.
    */
   void set_ans_b_pick_strategy(const strategy &s);

   /**
    * Sets the strategy for choosing which call to use if several exist.
    * \param s one of STRATEGY_FIRST, STRATEGY_LAST and STRATEGY_RANDOM.
    */
   void set_calls_pick_strategy(const strategy &s);

   /**
    * Sets the strategy for choosing which clause in the program to
    * use for EBC if several exist.
    * \param s one of STRATEGY_FIRST, STRATEGY_LAST and STRATEGY_RANDOM.
    */
   void set_possibles_strategy(const strategy &s);

   /**
    * Sets the percentage that local EBC is used.
    */
   void set_local_ebc_percentage(size_t p);

private:

   /**
    * Clears all remaining dynamic memory allocated by this model.
    */
   void deep_delete();

   /**
    * Gets an entry from the given set, using a given pick strategy.
    * \param s the set to pick from.
    * \param pick_strategy the strategy used.
    */
   entry get_next_entry(const std::set<entry> &s, compmod::strategy pick_strategy) const;

   /**
    * Uses the call picking strategy to find a call that can be used.
    * \return The call.
    */
   entry get_next_call() const;

   /**
    * Uses the answer alternative picking strategy to find an unused answer alternative.
    * \return The answer alternative.
    */
   entry get_next_answer_alternative() const;

   /**
    * Uses the answer bound picking strategy to find an unused answer bound.
    * \return The answer bound.
    */
   entry get_next_answer_bound() const;

   /**
    * Extends the model by EBC or EBCWA.
    */
   void extend_by_call();

   /**
    * Extends the model by EBAA.
    */
   void extend_by_answer_alternative();

   /**
    * Extends the model by EBAB.
    */
   void extend_by_answer_bound();

   /**
    * Adds a new entry to the model.
    * If the entry already exists in the model, the old and the new con_type
    * merge and replace the existing one.
    */
   std::pair<const c_rule*, unifier*> add_to_set(c_rule *rule);

   /**
    * Extends the model by applying an answer to the first body
    * literal of the given rule for the first time. This will not
    * be called if the answer is not the first one used on this
    * rule.
    * \param rule the rule whose first body literal unifies with the answer's head.
    * \param answer the answer to apply to the first body literal.
    */
   void extend_by_first_answer(const c_rule *rule, const con_type &answer);

   /**
    * Change the constraint C of this call into the disjunction of
    * C and another constraint. The change will be duefully propagated to
    * all answer extensions of this call, as to preserve
    * consistency. This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_disjunction_call(const c_rule *original, con_type addition);

   /**
    * Change the constraint C of this answer into the disjunction of
    * C and another constraint. This will add to the unused accumulated answer
    * alternative if it exists, otherwise add a new unused answer alternative.
    * This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_disjunction_answer(const c_rule *original, con_type addition);

   /**
    * Change the constraint C of this rule into the disjunction of
    * C and another constraint. It will call the appropriate specialised method.
    * This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_disjunction(const c_rule *original, con_type addition);

   /**
    * Change the constraint C of this call into the conjunction of
    * C and another constraint. The change will be duefully propagated to
    * all answer extensions of this call, as to preserve
    * consistency. This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_conjunction_call(const c_rule *original, con_type addition);

   /**
    * Change the constraint C of this answer into the conjunction of
    * C and another constraint. This will add to the unused accumulated answer
    * alternative if it exists, otherwise add a new unused answer alternative.
    * This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_conjunction_answer(const c_rule *original, con_type addition);

   /**
    * Change the constraint C of this rule into the conjunction of
    * C and another constraint. It will call the appropriate specialised method.
    * This is part of the update mechanism.
    * \param original the rule whose constraint to update.
    * \param addition the constraint to incorporate.
    */
   void change_constraint_conjunction(const c_rule *original, con_type addition);

   /**
    * Gets the answer (conjunction of alternative and bound) that
    * has been applied to the model for a given head, if one has
    * been applied.
    * \param t the head for which to search for an applied answer.
    * \param answer the answer (conjunction of alternative and bound) that
    * has been applied for this rule. The variables used are consistent with
    * those in the given head.
    * \return Truth value denoting if an answer was found or not. If the value
    * returned is false, neither the returned answer or unifier can be used.
    */
   bool get_answer(const c_term *t, con_type &answer) const;

   /**
    * Gets at least the calls (both active and obsolete) whose
    * first body literal unifies with the given literal. Both negated and
    * non-negated literals are returned.
    * \param t the literal to search for.
    * \param calls the vector that will contain the matching calls.
    */
   void get_calls(const c_term *t, std::vector<const c_rule*> &calls) const;

   /**
    * This method is called if a serious error occurs. This should not happen,
    * because it means that the model is inconsistent and cannot be trusted.
    * This also throws an compmod_error exception, which aborts the computation.
    */
   void fatal_error() const;

public:
   /**
    * The set of "active" calls, the calls which still has
    * possibilities for EBC or EBCWA.
    */
   std::set<entry> calls;

   /**
    * The set of obsolete calls, those that definitely can no
    * longer be used for EBC or EBCWA.
    */
   std::set<entry> ob_calls;

   /**
    * The set of disjunctions of answer alternatives not yet
    * used. There can be only one element per head variant.
    */
   std::set<entry> ans_a;

   /**
    * The set of conjunctions of answer bounds not yet used. There
    * can be only one element per head variant.
    */
   std::set<entry> ans_b;

   /**
    * The set of disjunctions of answer alternatives already
    * used. There can be only one element per head variant.
    */
   std::set<entry> ob_ans_a;

   /**
    * The set of conjunctions of answer bounds already used. There
    * can be only one element per head variant.
    */
   std::set<entry> ob_ans_b;

   /**
    * The strategy used for choosing which call to use.
    */
   strategy calls_pick_strategy;
      
   /**
    * The strategy used for choosing which answer alternative to
    * use.
    */
   strategy ans_a_pick_strategy;

   /**
    * The strategy used for choosing which answer bound to use.
    */
   strategy ans_b_pick_strategy;

   /**
    * The strategy used for choosing which clause in the program to use for EBC.
    */
   strategy possibles_strategy;

   /**
    * The global strategy used for choosing which entry type (answer
    * alternative/bound or call) to use to extend the model.
    */
   std::string pick_strategy;

   /**
    * The percent of EBC that should be done locally (derive the
    * constraint from the call ancestor).  The rest is done
    * globally (the derived call is not influenced by the parent
    * call).
    */
   size_t percent_local_extensions;

   /**
    * The answer extension of call entries, when they exist. This
    * is a function "Call(origin) -> Call(target) x Unifier", where
    * the unifier gives the unifier that unifies the origin and the
    * target.
    */
   std::map<const c_rule*, std::pair<const c_rule*, unifier*> > answer_extension;
      
   /**
    * A set of all calls that has been the origin of a EBC.
    */
   std::set<const c_rule*> used_ebc;

   /**
    * A vector of all the c_rule*:s used in the model, so that
    * final garbage collection can be used.
    */
   std::vector<c_rule*> all_rules;
};

#endif // COMPMOD_H
