#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include <map>
#include <vector>
#include <set>
#include <string>
#include <sstream>

#include <bdd.h>

/**
 * This is our constraint type. It uses BDDs from the BuDDy library
 * internally. All methods in this class needs to be defined for any
 * constraint type to be used in the program.
 */
class con_type
{
public:

   friend class conaux;

   /**
    * Write a text representation of this constraint to an ostream.
    * \param os the ostream to write to.
    * \param con the constraint to write.
    */
   friend std::ostream &operator<<(std::ostream &os, const con_type &con);

   /**
    * Creates a constraint representing the inclusion of every solution (true, default)
    * or the exclusion of every solution (false).
    * \param val the value denoting which extreme constraint to model.
    */
   con_type(bool val = true);

   /**
    * Creates a constraint representing a variable (will be created if it doesn't exist).
    * The solutions to this constraint will be those where the variable is solved.
    * \param var the name of the variable to model.
    */
   con_type(const std::string &var);

   /**
    * Returns the conjunction of two constraints. The solutions of the returned constraint
    * will be the intersection of the solutions to both constraints.
    * \param rhs the constraint on the right hand side of the expression.
    * \return The conjuncted constraint.
    */
   con_type operator&(const con_type &rhs) const;

   /**
    * Returns the disjunction of two constraints. The solutions of the returned constraint
    * will be the union of the solutions to both constraints.
    * \param rhs the constraint on the right hand side of the expression.
    * \return The disjuncted constraint.
    */
   con_type operator|(const con_type &rhs) const;

   /**
    * Returns the symmetric difference of two constraints. The solutions of the returned constraint
    * will be the solutions in one of the constraints, but not both.
    * \param rhs the constraint on the right hand side of the expression.
    * \return The XOR:ed constraint.
    */
   con_type operator^(const con_type &rhs) const;

   /**
    * Returns the complement of a constraint. The solutions of the returned constraint
    * will be those not solving the given constraint.
    * \return The complement constraint.
    */
   con_type operator!() const;

   /**
    * Returns true if this constraint has no solutions, false otherwise.
    * \return Truth value.
    */
   bool is_false() const;

   /**
    * Returns false if this constraint has no solutions, true otherwise.
    * \return Truth value.
    */
   bool is_true() const;

   /**
    * Projects this constraint onto the given variables.
    * \param to_keep the set of names of variables to keep.
    */
   void project(const std::set<std::string> &to_keep);

   /**
    * Gets the names of the constraint variables used by this
    * constraint.
    * \return The set of variable names. 
    */
   std::set<std::string> get_cvars();

private:

   /**
    * The internal BDD of this constraint.
    */
   bdd constraint;

   /**
    * The names of the variables used in this constraint.
    */
   std::set<std::string> cvars;
};

/**
 * This class holds static methods for manipulation constraints. Some
 * knowledge of the internal representation of the constraint class is used.
 */
class conaux
{
public:

   /**
    * This exception will be thrown in case of a malformed sat-expression in the program input.
    */
   class sat_exception: public std::exception {};
  
   /**
    * Initialise the constraint-related functionality. This method
    * must be called before any constraint activity takes place.
    */
   static void init(const std::string &constraint_settings);

   /**
    * Finalise the constraint-related functionality. This method
    * should be called to clean up the constraint biproducts.
    */
   static void finalise();

   /**
    * Gets the variable with the given name. If it does not exist, it is created first.
    * \param name the variable name.
    * \return The variable, in the internal constraint format.
    */
   static bdd get_variable(const std::string &name);

   /**
    * Gets the name of the variable with a given internal number.
    * \param num the number of the variable.
    * \return The variable name.
    */
   static std::string get_varname(const int num);

   /**
    * Resolves the constraint denoted by a sat-expression (from the program input).
    * \param t the tree of the sat-expression.
    * \return The resolved constraint.
    */
   static con_type resolve_sat(const class c_term *t);

   /**
    * Resolves a subtree of a sat-expression (from the program input).
    * \param term the subtree of the sat-expression (from the program input).
    * \param cvars the set of variables containing the variable names traversed so far.
    * \return The resolved subexpression.
    */
   static con_type build_tree(const class c_term *term, std::set<std::string> &cvars);

   /**
    * The function used for printing the solutions to a constraint. Used by BuDDy.
    * \param varset array of assigned values to the variables in a particular solution.
    * \param size the number of variables (size of varset)
    */
   static void allsathandler(char *varset, int size);

   /**
    * Reindexes the variables in the given constraint with the current global variable index.
    * This will change the variable names to include a suffix of "#<index>" and change
    * the underlying variables themselves accordingly.
    * \param con the constraint to reindex.
    */
   static void reindex(con_type &con);

   /**
    * Returns a biimplication between the given variables, effectively making them equal.
    * \param var1 the first variable name.
    * \param var2 the second variable name.
    * \return The constraint representing "<var1> <-> <var2>".
    */
   static con_type relate(const std::string &var1, const std::string &var2);

   /**
    * Report (all) the solutions of the given constraint to stdout.
    * \param con the constraint to report solutions to.
    */
   static void report_answer(const con_type &con);
      
   /**
    * Determine if the solutions of one constraint is a subset of the solutions to another constraint.
    * \param large the constraint expected to subsume the other.
    * \param small the constraint expected to have its solutions as a subset of the other's.
    * \return Truth value denoting whether the solutions of the second constraint is a subset of the
    * solutions to the first.
    */
   static bool subsumes(const con_type &large, const con_type &small);

private:
      
   /**
    * The number of variables allocated.
    */
   static int max_var_amount;

   /**
    * The mapping from variable name to variable.
    */
   static std::map<std::string, bdd> varmap;

   /**
    * The names of the variables, in order of creation.
    */
   static std::vector<std::string> varnames;
};

#endif // CONSTRAINT_H
