#ifndef PROLOG_CLASSS_H
#define PROLOG_CLASSS_H

#include <vector>
#include <string>
#include <set>

#include "stringsholder.hpp"
#include "unify.hpp"
#include "constraint.hpp"

// T_VARIABLE should always have the lowest value

/**
 * The types of terms there are.
 */
enum { T_VARIABLE = 1, T_FUNCTOR = 2, T_LIST = 3 };

/**
 * This class is the base class of all parser categories.
 */
class c_base
{
public:

   /**
    * Default constructor.
    */
   c_base();

   /**
    * Enables/disables the safety. When the safety is on, pointers
    * to all created c_* elements are stored away for garbage
    * collection later.
    * \param safe Safety value.
    */
   static void set_safety(const bool safe);

   /**
    * Garbage collects all c_* elements allocated during safe mode.
    */
   static void delete_us();

   /**
    * A pure virtual method that should delete a c_* instance and
    * all sub-c_* elements therein.
    */
   virtual void deep_delete() = 0;

   /**
    * Virtual destructor, needed because we have virtual methods.
    */
   virtual ~c_base();

private:

   /**
    * The c_* elements stored away for garbage collections.
    */
   static std::vector<c_base*> c_elements;

   /**
    * Safety mode flag.
    */
   static bool safety;
};

/**
 * A many-purpose element in the final result. It represents a functor
 * (head or body literal), a variable or a list depending on which
 * member variables are used.
 */
class c_term: public c_base
{
public:

   /**
    * Default constructor.
    */
   c_term();

   /**
    * Destructor.
    */
   ~c_term();

   /**
    * Calls delete on all subterms recursively. Note that this
    * should NOT be done on a term that wasn't created using
    * deep_copy during safety mode off.
    */
   virtual void deep_delete();

   /**
    * Copy this c_term recursively into another c_term.
    * \param copy allocated memory area for another term.
    * \param reindex should variables be reindexed?
    */
   void deep_copy(c_term *copy, bool reindex = false) const;
      
   /**
    * Inserts the variables of this term into a supplied set.
    * \param vars the set of variable names.
    */
   void get_variables(std::set<std::string> &vars) const;

   /**
    * Inserts the given list, starting at the given index, at this
    * position (only applicable in case this c_term is a list).
    * \param list all elements in the list to insert.
    * \param rest additional element to insert at the end (in case of a non-null-terminated list).
    * \param start_index the index to start looking up elements in the supplied list.
    */
   void insert_list(const std::vector<c_term*> &list,
                    const c_term *rest = 0,
                    const size_t start_index = 0);

   /**
    * Mark this term as sorted before another term.
    * \param t the other term.
    */
   void sort_before(const c_term *t) const;

   /**
    * Mark this term as sorted at the same place as another term.
    * \param t the other term.
    */
   void sort_same(const c_term *t) const;

   /**
    * The number of c_* instances allocated and not destroyed.
    */
   static int count;

   /**
    * The name in case of a functor or variable.
    */
   std::string name;

   /**
    * The arguments in case of a functor.
    */
   std::vector<c_term*> arglist; 

   /**
    * Is this a negated functor?
    */
   bool negated;

   /**
    * Head of the list in case of a list, 0 otherwise.
    */
   c_term *list_head; 

   /**
    * Tail of the list in case of a list, 0 otherwise.
    */
   c_term *list_tail;

   /**
    * Is this an empty list?
    */
   bool list_empty;

   /**
    * This term was given after a pipe in a list and therefore
    * matches the rest of any list
    */
   bool match_rest; 

   /**
    * The type of this term. (T_VARIABLE, T_FUNCTOR or T_LIST)       
    */
   int type;

   /**
    * Cached orderings between this term and others.
    */
   mutable std::map<const c_term*, int> sorted_terms;
};

/**
 * This class is the base for all AST nodes used during parsing.
 */
class n_base
{
public:

   /**
    * Default constructor.
    */
   n_base();

   /**
    * The number of n_* instances allocated and not destroyed.
    */
   static int count;

   /**
    * Destructor.
    */
   ~n_base();

   /**
    * Garbage collect the memory for n_* nodes allocated during parsing.
    */
   static void delete_us();

   /**
    * The n_* nodes stored away during parsing for garbage collection.
    */
   static std::vector<n_base*> nodes;
};

/**
 * This class represents an argument to a functor or an element of a list.
 */
class n_arg: public n_base
{
public:

   /**
    * Constructor.
    */
   n_arg();

   /**
    * The term that is the argument.
    */
   c_term *term;
};

/**
 * This class represents a list of arguments to a functor or list of elements to a list.
 */
class n_arglist: public n_base
{
public:

   /**
    * Constructor taking the car and cdr of a linked list of arguments/elements.
    */
   n_arglist(const n_arg*, const n_arglist*);

   /**
    * The list of arguments/elements in the list from this point.
    */
   std::vector<c_term*> terms;
};

/**
 * This class represents a functor.
 */
class n_functor: public n_arg
{
public:

   /**
    * Constructor taking the functor name and the list of arguments.
    */
   n_functor(const std::string, const n_arglist*);

   /**
    * Negate this functor. Two calls to this method removes negation.
    */
   void negate();
};

/**
 * This class represents a variable.
 */
class n_variable: public n_arg
{
public:

   /**
    * Constructor taking the variable name.
    */
   n_variable(const std::string);
};

/**
 * This class represents a list.
 */
class n_list: public n_arg
{
public:

   /**
    * Functor taking list of elements.
    */
   n_list(const n_arglist*);
      
   /**
    * Functor taking a list of elements and a final element (in case of non-null-terminated lists).
    */
   n_list(const n_arglist*, const n_arg*);

   /**
    * Default constructor, creates an empty list.
    */
   n_list();
};

/**
 * This class represents a CLP rule, holding a head, an associated
 * constraint and zero or more body literals.
 */
class c_rule: public c_base
{
public:

   /**
    * Default constructor. Creates a head.
    */
   c_rule();
            
   /**
    * Destructor.
    */
   ~c_rule();

   /**
    * Calls delete on the head and body recursively. Note that this
    * should NOT be done when the rule contains term that wasn't
    * created using deep_copy during safety mode off.
    */
   void deep_delete();

   /**
    * Copy this c_rule recursively into another c_rule.
    * \param copy allocated memory area for another rule.
    * \param reindex should variables be reindexed?
    */
   void deep_copy(c_rule *copy, bool reindex = false) const;

   /**
    * Project the constraint in this rule to the variables used
    * somewhere in the rule.
    */
   void project_constraint() const;

   /**
    * Inserts the variables names used in this rule into the
    * supplied set of strings.
    */
   void get_variables(std::set<std::string> &vars) const;

   /**
    * Mark this rule as sorted before another rule.
    * \param r the other rule.
    */
   void sort_before(const c_rule *r) const;

   /**
    * Mark this rule as sorted at the same place as another rule.
    * \param r the other rule.
    */
   void sort_same(const c_rule *r) const;

   /**
    * The index for variables in a new rule instance,
    * increased by 1 for every rule instance.
    */
   static int global_variable_index; 

   /**
    * The number of c_rule instances allocated and not destroyed.
    */
   static int count;

   /**
    * The head of this rule, is of type T_FUNCTOR and isn't negated.
    */
   c_term *head;

   /**
    * The body literals of this rule (none if it is an answer).
    * All of these literals are of type T_FUNCTOR.
    */
   std::vector<c_term*> body;
      
   /**
    * Cached orderings between this rule and others.
    */
   mutable std::map<const c_rule*, int> sorted_rules;

   /**
    * The constraint of this rule.
    */
   mutable con_type constraint;
};

/**
 * This class represents the body of a rule.
 */
class n_body: public n_base
{
public:

   /**
    * Functor taking the car and cdr of the body.
    */
   n_body(const n_functor*, const n_body*);

   /**
    * The functors that constitute the body.
    */
   std::vector<c_term*> functors;

   /**
    * Constraint used during parsing (will be something if we have
    * a sat-expression in the program input for this rule).
    */
   static con_type constraint;
};

/**
 * This class represents a rule.
 */
class n_rule: public n_base
{
public:

   /**
    * Constructor taking the head and the body of the rule (in case of a call).
    */
   n_rule(const n_functor*, const n_body*);

   /**
    * Constructor taking the head of the rule (in case of an answer).
    */
   n_rule(const n_functor*);
      
   /**
    * The c_rule representing this rule.
    */
   c_rule *rule;
};

/**
 * This class represents a linked list of rules.
 */
class n_rules: public n_base
{
public:

   /**
    * Constructor taking the car and cdr of the rule list.
    */
   n_rules(const n_rule*, const n_rules*);

   /**
    * The c_rule representation of the rules in this list.
    */
   std::vector<c_rule*> rules;
};

/**
 * This class represents the final representation of the full program parsed from the input.
 */
class prolog_program
{ 
public:

   /**
    * Constructor taking the rules that will be copied.
    * \param rs the CLP rules constituting the program.
    */
   prolog_program(const std::vector<c_rule*> &rs);

   /**
    * The stored rules.
    */
   std::vector<c_rule*> rules;
};

/**
 * This class acts as sorting functions for c_terms and c_rules.
 */
class c_sorter
{
public:

   /**
    * A sorting function for c_rules. Answers and calls may not be
    * mixed in the same sorting.
    *
    * This function merely calls c_term_order to check if the order
    * was "c1 before c2".
    *
    * \param c1 the left hand rule.
    * \param c2 the right hand rule.
    * \return Is c1 sorted before c2?
    */
   static bool c_rule_lt(const c_rule *c1, const c_rule *c2);

   /**
    * A sorting function for c_terms.
    *
    * This function merely calls c_term_order to check if the order
    * was "c1 before c2".
    *
    * \param c1 the left hand term
    * \param c2 the right hand term
    * \return Is c1 sorted before c2?
    */
   static bool c_term_lt(const c_term *c1, const c_term *c2);

private:

   /**
    * Determines the order of two terms.
    *
    * The terms are first sorted in the order "variables before
    * functors before lists". After that, they are sorted on (in case
    * of functors) primarily on functor name, secondarily recursively
    * on arguments and tertially on negatedness. In case of lists, the
    * terms are sorted recursively on the list elements, shortest list
    * first. In case of variables no sorting is done (two variables
    * are always equal, resulting in that only one variant of a term
    * may exist).
    *
    * \param c1 the left hand term.
    * \param c2 the right hand term.
    * \return -1 iff c1 < c2, 0 iff c1 = c2 or 1 iff c1 > c2 
    */
   static int c_term_order(const c_term *c1, const c_term *c2);

   /**
    * Determines the order of two rules.
    *
    * In case of two answers, the one with the least head (by
    * c_term_lt) is preferred.
    *
    * In case of two calls, they are primarily sorted recursively
    * on body literals, secondarily on body length and tertiarily
    * on the head. All term sorting is done through c_term_lt.
    *
    * \param c1 the left hand rule.
    * \param c2 the right hand rule.
    * \return -1 iff c1 < c2, 0 iff c1 = c2 or 1 iff c1 > c2 
    */
   static int c_rule_order(const c_rule *c1, const c_rule *c2);
};

/**
 * Write a text representation of this program to an ostream.
 * \param os the ostream to write to.
 * \param p the program to write.
 */
std::ostream &operator<<(std::ostream &os, const prolog_program *p);

/**
 * Write a text representation of this rule to an ostream.
 * \param os the ostream to write to.
 * \param r the rule to write.
 */
std::ostream &operator<<(std::ostream &os, const c_rule *r);

/**
 * Write a text representation of this term to an ostream.
 * \param os the ostream to write to.
 * \param t the term to write.
 */
std::ostream &operator<<(std::ostream &os, const c_term *t);


#endif // PARSER_CLASSES
