#include "unify.hpp"
#include "log.hpp"

using namespace std;

int unifier::count = 0;

unifier::unifier()
{
   count++;
}

unifier::~unifier()
{
   count--;

   map<string, c_term*>::iterator it = bindings.begin();
   while(it != bindings.end())
   { 
      it->second->deep_delete();
      ++it;
   }
}

void unifier::apply_to(con_type &con) const
{
   if(bindings.size() == 0) return;

   set<string> cvars = con.get_cvars();
   map<string, c_term*>::const_iterator mapit = bindings.begin();
   set<string> to_remove;
   set<string> to_add;

   while(mapit != bindings.end())
   {
      if(mapit->second->type != T_VARIABLE)
      {
	 ++mapit;
	 continue;
      }

      if(cvars.find(mapit->first) == cvars.end()
	 && cvars.find(mapit->second->name) == cvars.end())
      {
	 ++mapit;
	 continue;
      }

      
      to_remove.insert(mapit->first);
      to_add.insert(mapit->second->name);

      // Make the two variables equal
      con = con & conaux::relate(mapit->first, mapit->second->name);
      ++mapit;
   }

   set_difference(cvars.begin(), cvars.end(), to_remove.begin(), to_remove.end(), inserter(to_add, to_add.end()));
   con.project(to_add);
}

void unifier::apply_to(class c_rule *r) const
{
   if(bindings.size() == 0) return;
   
   lout(L_DBG) << "Applying " << this << " to " << r << endl;
   
   // Apply to the head
   apply_to(r->head);

   // Apply to the constraint
   apply_to(r->constraint);

   // Apply to the body
   for(size_t i = 0; i < r->body.size(); i++)
      apply_to(r->body[i]);
}

void unifier::apply_to(class c_term *t) const
{
   if(bindings.size() == 0) return;

   switch(t->type)
   {
      case T_VARIABLE:
      {
	 map<string, c_term*>::const_iterator it = bindings.find(t->name);
	 if(it != bindings.end())
	    it->second->deep_copy(t);
      }
      break;

      case T_FUNCTOR:
      {
	 for(size_t i = 0; i < t->arglist.size(); i++)
	    apply_to(t->arglist[i]);
      }
      break;

      case T_LIST:
      {
	 if(t->list_head != 0) apply_to(t->list_head);
	 if(t->list_tail != 0) apply_to(t->list_tail);
      }
      break;

      default:
	 break;
   }
}

bool unifier::add_binding(const string &s, const c_term *t)
{
   // Do not add bindings of the kind X=X
   if(t->type == T_VARIABLE && s == t->name)
      return true;

   c_term *temp;
   
   map<string, c_term*>::iterator it = bindings.find(s);
   if(it != bindings.end())
   {
      c_term *old = it->second;
      if(!unify(old, t))
	 return false;
      temp = new c_term;
      t->deep_copy(temp);
      apply_to(temp);
      delete old;
   }
   else
   {
      temp = new c_term;
      t->deep_copy(temp);
   }

   bindings[s] = temp;

   return true;
}

bool unifier::unify(const class c_term *t1, const class c_term *t2)
{
   // In the improbable case that the pointers are equal
   if(t1 == t2)
      return true;

   if(t1->type == T_VARIABLE && t2->type == T_VARIABLE)
   {
      return add_binding(t2->name, t1);
   }
   else if(t1->type == T_VARIABLE)
   {
      return add_binding(t1->name, t2);
   }
   else if(t2->type == T_VARIABLE)
   {
      return add_binding(t2->name, t1);
   }
   else if(t1->type != t2->type) // Different types
   {
      return false;
   }
   else if(t1->type == T_FUNCTOR) // Two functors
   {
      // Different functor names
      if(t1->name != t2->name)
	 return false;
      
      // Arity is always the same if they have the same name

      // Unify the arguments
      for(size_t i = 0; i < t1->arglist.size(); i++)
      {
	 if(!unify(t1->arglist[i], t2->arglist[i]))
	    return false;
      }
      
      return true;
   }
   else if(t1->type == T_LIST) // Two lists
   {
      const c_term *temp1 = t1,
	 *temp2 = t2;
      
      while(true)
      {
	 // After-pipe-expression (matches rest of list)
	 if(temp1->match_rest || temp2->match_rest)
	 {
	    return unify(temp1, temp2);
	 }

	 // End-of-list
	 if(temp1->list_empty)
	 {
	    return temp2->list_empty;
	 }
	 else if(temp2->list_empty)
	 {
	    return temp1->list_empty;
	 }
	 
	 // Normal list element
	 if(!unify(temp1->list_head, temp2->list_head))
	 {
	    return false;
	 }

	 temp1 = temp1->list_tail;
	 temp2 = temp2->list_tail;
      }
   }

   return false;
}

bool unifier::unify(const class c_rule *r1, const class c_rule *r2)
{
   if(r1->body.size() != r2->body.size())
      return false;
   
   if(!unify(r1->head, r2->head))
      return false;

   for(size_t i = 0; i < r1->body.size(); i++)
   {
      if(!unify(r1->body[i], r2->body[i]))
	 return false;
   }

   return true;
}

// Prints all the bindings
ostream &operator<<(ostream &os, const unifier *u)
{
   os << "{ ";

   map<string, c_term*>::const_iterator mapit = u->bindings.begin();
   
   if(mapit != u->bindings.end())
   {
      os << mapit->first << "=" << mapit->second;
      ++mapit;
   }

   while(mapit != u->bindings.end())
   {
      os << ", " << mapit->first << "=" << mapit->second;
      ++mapit;
   }

   os << " }";

   return os;
}
