#ifndef UNIFY_H
#define UNIFY_H

#include <map>
#include <set>
#include <string>

#include "parser-classes.hpp"
#include "constraint.hpp"

class unifier
{
public:

   /**
    * Write a text representation of this unifier to an ostream.
    * \param os the ostream to write to.
    * \param u the unifier to write.
    */
   friend std::ostream &operator<<(std::ostream &os, const unifier *u);

   /**
    * Destructor that deeply deletes all terms in its bindings.
    */
   ~unifier();
      
   /**
    * Default constructor.
    */
   unifier();

   /**
    * Adds a binding for a variable name to a term. The given term is copied.
    * \param s the variable name.
    * \param t the term.
    * \return Could the binding be added? (did it not collide with earlier non-unifiable bindings?)
    */
   bool add_binding(const std::string &s, const c_term *t);

   /**
    * Apply this binding to a term.
    * \param t the term to apply the unifier to.
    */
   void apply_to(class c_term *t) const;

   /**
    * Apply this binding to a rule.
    * \param r the rule to apply the unifier to.
    */
   void apply_to(class c_rule *r) const;

   /**
    * Apply this binding to a constraint.
    * \param c the constraint to apply the unifier to.
    */
   void apply_to(con_type &c) const;

   /**
    * Get the most general unifier of two terms. The necessary
    * bindings will be added to the unifier. If two variables unify
    * with eachother, the binding is directed from the second term
    * to the first.
    * \param t1 the first term.
    * \param t2 the second term.
    * \return Do they unify? If this is false, the unifier must not be used.
    */
   bool unify(const class c_term *t1, const class c_term *t2);

   /**
    * Get the most general unifier of two rules. The necessary
    * bindings will be added to the unifier. If two variables unify
    * with eachother, the binding is directed from the second term
    * to the first.
    * \param r1 the first rule.
    * \param r2 the second rule.
    * \return Do they unify? If this is false, the unifier must not be used.
    */
   bool unify(const class c_rule *r1, const class c_rule *r2);

   /**
    * The number of unifiers allocated and not destroyed.
    */
   static int count;

private:

   std::map<std::string, c_term*> bindings;
};

#endif //UNIFY_H
